package com.example.android.constantesvitales;

import android.annotation.SuppressLint;
import android.content.Context;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.BroadcastReceiver;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Build;
import android.support.annotation.RequiresApi;
import android.support.v7.app.AppCompatActivity;
import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;
import org.jetbrains.annotations.NotNull;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import android.os.Handler;
import android.os.Message;
import java.util.ArrayList;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicBoolean;


@SuppressWarnings("serial")
public class ActividadConstantes extends AppCompatActivity {

    TextView conectarTextView;
    TextView nombreTextView;
    TextView edadTextView;
    TextView alturaTextView;
    TextView pesoTextView;
    ListView listaConstantesView;
    LinearLayout datos_perfil;
    LinearLayout detalles_perfil;
    ImageView flecha;


    ArrayList<ConstanteVital> listaConstantesVitales;
    private BluetoothAdapter btAdapter = BluetoothAdapter.getDefaultAdapter();
    private static final int REQUEST_ENABLE_BT = 1;
    private final AtomicBoolean estadoConexion = new AtomicBoolean(false);
    private final AtomicBoolean botonConectar = new AtomicBoolean(true);
    Set<BluetoothDevice> listaDispositivos = null;
    private static final String direccion = "00:02:5B:00:A0:18";
    private static final String MENSAJE_PREDETERMINADO =
            "{\"datos\":{\"ppg\":0,\"oxigeno\":0,\"pulso\":0}}";
    public BluetoothSocket btSocket = null;
    private static final UUID MI_UUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB");

    Paciente paciente = new Paciente("",-1,-1,-1);
    Handler handlerConstantes;
    Handler handlerEstadoConexion;


    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.actividad_constantes);
        inicializarViews();

        handlerConstantes = new Handler() {
            @Override
            public void handleMessage(Message msg) {

                Bundle bundle = msg.getData();
                String json = bundle.getString("llaveConstantes");

                listaConstantesVitales = ObtencionDatos.getListaConstantes(paciente,
                        json, getResources());
                ConstantesAdapter adapter = new ConstantesAdapter(getApplicationContext(),
                        listaConstantesVitales, getResources());
                listaConstantesView = (ListView) findViewById(R.id.lista);
                listaConstantesView.setAdapter(adapter);

            }
        };

        handlerEstadoConexion = new Handler() {
            @Override
            public void handleMessage(Message msg) {
                Bundle bundle = msg.getData();
                String estado = bundle.getString("llaveConexion");
                if(estado==getString(R.string.conectar)){
                    Toast.makeText(getBaseContext(), getString(R.string.no_compatibles),
                            Toast.LENGTH_LONG).show();
                }
                conectarTextView.setText(estado);
                botonConectar.set(true);                            //Volvemos a habilitar el botón
            }
        };
    }


    @Override
    protected void onActivityResult (int requestCode, int resultCode, Intent data) {
        switch(requestCode)
        {
            case REQUEST_ENABLE_BT:
            {
                if(resultCode == RESULT_OK)
                {
                    Toast.makeText(getBaseContext(), getText((R.string.buscando)),
                            Toast.LENGTH_LONG).show();
                    botonConectar.set(false);
                    conectarTextView.setText(getString(R.string.conectando));
                    EstablecerConexion hiloEstablecerConexion = new EstablecerConexion();
                    hiloEstablecerConexion.start();
                }
                else
                {
                    Toast.makeText(this, getString(R.string.peticion_cancelada),
                            Toast.LENGTH_SHORT).show();
                }
                break;
            }

            default:
                break;
        }
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
        this.unregisterReceiver(broadcastReceiver);

        if (btSocket!=null) {
            try {
                btSocket.close();
                estadoConexion.set(false);
            }
            catch (IOException e){
                Log.e("Cierre de socket", "Error al cerrar el socket.", e);
            }
        }
    }


    /* Inicializa todas las vistas y listeners */

    private void inicializarViews() {

        /*Intentamos cargar los datos del paciente creados en la actividad inicial y usamos el
        adapter para mostrarlos en la listView*/

        try{
            paciente = (Paciente) getIntent().getExtras().getSerializable("paciente");
            rellenaPerfil(paciente);

            listaConstantesVitales = ObtencionDatos.getListaConstantes(paciente,
                    MENSAJE_PREDETERMINADO, getResources());
            ConstantesAdapter adapter = new ConstantesAdapter(this, listaConstantesVitales,
                    getResources());
            listaConstantesView = (ListView) findViewById(R.id.lista);
            listaConstantesView.setAdapter(adapter);

        }
        catch (Error e){
            Log.e("Cargar lista", "Error al cargar la lista de datos", e);
        }


        /*Creamos un listener en la vista del perfil que permitira desplegar y contraer los datos */

        conectarTextView = (TextView)findViewById(R.id.conectar);
        datos_perfil = (LinearLayout) findViewById(R.id.datos_perfil);
        detalles_perfil = (LinearLayout) findViewById(R.id.detalles_perfil);
        flecha = (ImageView) findViewById(R.id.flecha);
        detalles_perfil.setVisibility(View.GONE);
        flecha.setBackgroundResource(R.drawable.flecha_expandir);

        datos_perfil.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                if(detalles_perfil.getVisibility()==View.GONE) {
                    detalles_perfil.setVisibility(View.VISIBLE);
                    flecha.setBackgroundResource(R.drawable.flecha_contraer);
                }
                else {
                    detalles_perfil.setVisibility(View.GONE);
                    flecha.setBackgroundResource(R.drawable.flecha_expandir);
                }
            }
        });


        /* Comprobamos si nuestro dispositivo dispone de Bluetooth. Si no es así, el botón para
        conectarse al mismo permanecerá inhabilitado */

        if(btAdapter == null)
        {
            conectarTextView.setBackgroundResource(R.color.colorDissabled);
            conectarTextView.setText(getString(R.string.BT_no_disponible));
        }

        else {
            registrarBroadcastReceiver();
            conectarTextView.setBackgroundResource(R.color.colorPrimary);

            /* Comprobamos si el Bluetooth está activo o no y cambiamos el texto acorde a ello */

            if(estadoConexion.get()) {
                conectarTextView.setText(getString(R.string.desconectar));
            }
            else {
                conectarTextView.setText(getString(R.string.conectar));
            }

            /* Establecemos la función del botón para conectarnos o desconectarnos */

            conectarTextView.setOnClickListener(new View.OnClickListener() {
                @Override
                public void onClick(View pulsador) {
                    if (botonConectar.get()) {                        //Si está habilitado el botón
                        try {
                            if (!(estadoConexion.get())) {            //Si el hilo no está activo
                                if (!btAdapter.isEnabled()) {
                                    conectarBt(btAdapter);
                                } else {
                                    Toast.makeText(getBaseContext(), getText((R.string.buscando)),
                                            Toast.LENGTH_LONG).show();
                                    botonConectar.set(false);             //Deshabilitamos el botón
                                    conectarTextView.setText(getString(R.string.conectando));
                                    EstablecerConexion hiloEstablecerConexion =
                                            new EstablecerConexion();
                                    hiloEstablecerConexion.start();
                                }
                            } else {                                      //Si el hilo está activo
                                estadoConexion.set(false);
                                conectarTextView.setText(getString(R.string.conectar));

                                listaConstantesVitales = ObtencionDatos.getListaConstantes(paciente,
                                        MENSAJE_PREDETERMINADO, getResources());
                                ConstantesAdapter adapter = new ConstantesAdapter
                                        (getApplicationContext(), listaConstantesVitales,
                                                getResources());
                                listaConstantesView = (ListView) findViewById(R.id.lista);
                                listaConstantesView.setAdapter(adapter);
                            }
                        } catch (Error e) {
                            e.printStackTrace();
                        }
                    }
                }
            });
        }
    }


    /* Carga los datos del paciente y crea la vista del perfil, con la opcion de contraer y
    expandir los detalles */

    @SuppressLint("SetTextI18n")
    private void rellenaPerfil(@NotNull Paciente paciente){

        /* Tomamos los valores de la clase paciente con los datos que hemos tomado en la actividad
        del inicio y establecemos los valores de los text views correspondientes */

        nombreTextView = (TextView) findViewById(R.id.nombre);
        nombreTextView.setText(paciente.getNombre());

        edadTextView = (TextView) findViewById(R.id.edad);
        int edad = paciente.getEdad();
        if(edad==-1){
            edadTextView.setText("--");
        }
        else {
            edadTextView.setText(Integer.toString(edad));
        }

        alturaTextView = (TextView) findViewById(R.id.altura);
        int altura = paciente.getAltura();
        if(altura==-1){
            alturaTextView.setText("--");
        }
        else{
            alturaTextView.setText(Integer.toString(altura));
        }

        pesoTextView = (TextView) findViewById(R.id.peso);
        int peso = paciente.getPeso();
        if(peso==-1){
            pesoTextView.setText("--");
        }
        else{
            pesoTextView.setText(Integer.toString(peso));
        }
    }


    /*Conecta o desconecta el Bluetooth*/

    private  void conectarBt(@NotNull BluetoothAdapter btAdapter) {

        /* Si el dispositivo esta apagado entonces lo encendemos, buscamos dispositivos compatibles,
        y nos conectamos de ser posible */

        if (!(btAdapter.isEnabled())) {

            try {
                /* Pedimos permisos al usuario para conectar el Bluetooth */
                Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
                startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT);
            } catch (Error e) {
                Log.e("Cargar lista", "Error al cargar la lista de datos", e);
            }
        }
    }


    /* Registramos el broadcast receiver */

    private void registrarBroadcastReceiver() {

        IntentFilter filtro = new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED);
        filtro.addAction(BluetoothDevice.ACTION_FOUND);
        this.registerReceiver(broadcastReceiver, filtro);
    }


    /*Definimos el BroadcastReceiver para detectar cambios en el adaptador Bluetooth*/

    private final BroadcastReceiver broadcastReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, @NotNull Intent intent) {
            final String accion = intent.getAction();
            final int estado = intent.getIntExtra(BluetoothAdapter.EXTRA_STATE,
                    BluetoothAdapter.ERROR);

            if (BluetoothAdapter.ACTION_STATE_CHANGED.equals(accion)) {
                switch (estado) {
                    case BluetoothAdapter.STATE_OFF: {
                        estadoConexion.set(false);
                        ((TextView)findViewById(R.id.conectar)).setText(getString(R.string.conectar));
                        break;
                    }

                    default:
                        break;
                }
            }

            if(BluetoothDevice.ACTION_FOUND.equals(accion)) {
                BluetoothDevice device = (BluetoothDevice) intent.getParcelableExtra
                        (BluetoothDevice.EXTRA_DEVICE);

                Toast.makeText(getApplicationContext(), getString(R.string.encontrado) +
                                device.getName(),
                        Toast.LENGTH_SHORT).show();
            }
        }
    };



    /*Hilo del proceso para recibir los mensajes que llegan a traves del socket de conexión*/

    private class HiloTransmisionDatos extends Thread{

        private BluetoothSocket mSocket;
        private InputStream mFlujoEntrada;
        BufferedReader br;
        final byte delimitador = 10;           /*El caracter delimitador es el correspondiente
                                               al salto de línea \n en código ASCII, esto es 10*/
        private static final char DELIMITADOR = '\n';
        String mensaje  = "";

        HiloTransmisionDatos(BluetoothSocket socket){
            mSocket = socket;
        }

        private String leer(){
            String s = "";

            try {
                byte[] buffer = new byte[1024];
                int bytes = mFlujoEntrada.read(buffer);
                s = new String(buffer, "ASCII");
                s = s.substring(0, bytes);
            }catch (IOException e){

            }
            return s;
        }

        private void enviaHAndlerConstantes(String s) {

            Message msg = Message.obtain();
            Bundle bundle = new Bundle();
            bundle.putString("llaveConstantes", s);
            msg.setData(bundle);
            handlerConstantes.sendMessage(msg);
        }

        private void construirMensaje() {

            //Buscamos el primer delimitador que haya en el buffer
            int indice = mensaje.indexOf(DELIMITADOR);

            //Si no hay ninguno salimos
            if (indice == -1) {
                return;
            }

            //Construimos el mensaje completo
            String s = mensaje.substring(0, indice);

            //Eliminamos el mensaje del buffer
            mensaje = mensaje.substring(indice + 1);

            //Si empieza correctamente es que el json esta entero y lo enviamos
            if(s.contains("{\"datos\":")){
                enviaHAndlerConstantes(s);
            }

            //Seguimos buscando
            construirMensaje();
        }


        @RequiresApi(api = Build.VERSION_CODES.N)
        public void run() {

            estadoConexion.set(true);

            try{
                mFlujoEntrada = mSocket.getInputStream();
            }
            catch(IOException e){
                Log.e("Flujo de entrada", "Error al crear el flujo de entrada", e);
            }

            while (estadoConexion.get() && !this.isInterrupted() && btSocket.isConnected()) {

                if (mFlujoEntrada == null){
                    break;
                }

                String s = leer();

                if (s.length() > 0)
                    mensaje += s;

                construirMensaje();

            }
            if(!btSocket.isConnected() || !estadoConexion.get()){
                ((TextView)findViewById(R.id.conectar)).setText(getString(R.string.conectar));
                estadoConexion.set(false);
                if (mFlujoEntrada != null) {
                    try {mFlujoEntrada.close();} catch (Exception e) { e.printStackTrace(); }
                }
            }
        }
    }



    private class EstablecerConexion extends Thread {


        EstablecerConexion(){
        }

        public void run() {

            Message msg = handlerEstadoConexion.obtainMessage();
            Bundle bundle = new Bundle();

            /*Cada vez que llamemos a la función, vaciaremos el set de dispositivos para volver a
            buscarlos, en caso de que el set esté vacío*/

            if (listaDispositivos != null) {        //Comprobamos si es null antes de ver su tamaño
                if (listaDispositivos.size() > 0) {
                    for (BluetoothDevice dispositivo : listaDispositivos) {
                        try {
                            Method m = dispositivo.getClass().getMethod("eliminarVinculos",
                                    (Class[]) null);
                            m.invoke(dispositivo, (Object[]) null);

                        } catch (Exception e) {
                            Log.e("Vaciado de lista", "Error en el vaciado de la lista.");
                        }
                    }
                }
            }

            /*Si ya hay un socket creado, lo cerramos*/

            if (btSocket != null) {
                try {
                    btSocket.close();
                } catch (IOException e) {
                    Log.e("Cierre de socket", "Error al cerrar el socket.", e);
                }
            }

            /*Buscamos dispositivos asociados y llenamos el set de dispositivos que hemos
            encontrado*/

            listaDispositivos = btAdapter.getBondedDevices();

            if(listaDispositivos!=null) {
                for (BluetoothDevice dispositivo : listaDispositivos) {
                    if (dispositivo.getAddress().equals(direccion)) {
                        try {
                            btSocket = creaBluetoothSocket(dispositivo);
                        }catch (IOException e) {
                            Toast.makeText(getBaseContext(), getString(R.string.error_socket),
                                    Toast.LENGTH_LONG).show();
                        }
                        break;
                    }
                }

                /*Nos conectamos al socket. Este proceso tarda unos segundos bloqueando al hilo
                principal, y por eso debe ejecutarse en otro hilo*/

                if (btSocket != null) {
                    try {
                        btSocket.connect();
                        Log.d("Conectado", "Conectado: " + btSocket.isConnected());
                    } catch (IOException e) {
                        try {
                            btSocket.close();
                            //ConexionFallida(bundle, msg);
                        } catch (IOException e2) {
                            Log.e("Cierre de socket", "Error al cerrar el socket.", e2);
                        }
                    }

                    if (btSocket.isConnected()) {
                        HiloTransmisionDatos hiloTransmisionDatos = new HiloTransmisionDatos
                                (btSocket);
                        hiloTransmisionDatos.start();
                        bundle.putString("llaveConexion", getString(R.string.desconectar));
                        msg.setData(bundle);
                        handlerEstadoConexion.sendMessage(msg);
                    }
                    else {
                        ConexionFallida(bundle, msg);
                    }
                }else {
                    ConexionFallida(bundle, msg);
                }
            }else {
                ConexionFallida(bundle, msg);
            }
        }
    }



    private BluetoothSocket creaBluetoothSocket(BluetoothDevice dispositivo) throws IOException {

        return dispositivo.createRfcommSocketToServiceRecord(MI_UUID);
    }

    private void ConexionFallida(Bundle bundle, Message msg){
        bundle.putString("llaveConexion", getString(R.string.conectar));
        estadoConexion.set(false);
        msg.setData(bundle);
        handlerEstadoConexion.sendMessage(msg);
    }
}

